import torch
import numpy as np

class Linear(nn.Module):
    def __init__(self, input_features, output_features, bias=True):
        super(Linear, self).__init__()
        self.input_features = input_features
        self.output_features = output_features

        # nn.Parameter 能够将 Torch Tensor 注册到定义的模块中
        # 从而能够实现 
        # 1. 跟随模块在 CPU / GPU 之间的转移
        # 2. 在执行 module.parameters() 时返回该矩阵
        #    从而方便优化器对其反向传播
        # 3. 默认该 Tensor 被设置为可反传 (预留了梯度存储空间)
        self.weight = nn.Parameter(
                            torch.Tensor(output_features, input_features)
                       )
        if bias:
            self.bias = nn.Parameter(torch.Tensor(output_features))

        # 可为参数赋初始值，但应当用更好的初始值设置方法
        self.weight.data.uniform_(-0.1, 0.1)
        if self.bias is not None:
            self.bias.data.uniform_(-0.1, 0.1)

    def forward(self, input):
        # 执行实际的操作
        # input 形状是 (N, dim)，表示 N 个向量
        # self.weights 形状是 (out_dim, im_dim)
        # 需要对 N 个向量中的每一个做矩阵向量乘法
        # 等价于下式
        return input.mm(self.weight.t()) + self.bias

class LinearFunction(torch.autograd.Function):

    # 对 forward 和 backward 方法需要用 staticmethod 修饰
    @staticmethod
    def forward(ctx, input, weight, bias=None):
        #在反向传播时需要用到当前层的输出
        # 所以需要在计算图上将数值存下来
        ctx.save_for_backward(input, weight, bias)
        output = input.mm(weight.t())
        if bias is not None:
            output += bias.unsqueeze(0).expand_as(output)
        return output

    # 该操作仅产生一个输出，因此在反向传播时仅存在一个上层传来的梯度
    @staticmethod
    def backward(ctx, grad_output):
        # 提取出之前前向传播保存下来的矩阵
        input, weight, bias = ctx.saved_tensors
        grad_input = grad_weight = grad_bias = None

        # 检查各个参数/输入是否需要梯度传递
        # 若需要，则按照公式分别计算反向传播过程中
        # 作用于 W 和内部偏置 b 的梯度
        if ctx.needs_input_grad[0]:
            grad_input = grad_output.mm(weight)
        if ctx.needs_input_grad[1]:
            grad_weight = grad_output.t().mm(input)
        if bias is not None and ctx.needs_input_grad[2]:
            grad_bias = grad_output.sum(0)

        return grad_input, grad_weight, grad_bias
# 此后 linear 就是一个可以被直接使用的操作 (op)
linear = LinearFunction.apply
